/*
 * blu-save helps you to backup your bluray discs. This is the gtk version.
 * Copyright 2019 Yunxiang Li
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "bluray.h"
#include "file/file.h"

#include <glib.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#define DEFAULT_OUT "out"

typedef struct {
    const char **out_paths;
    size_t size;
    size_t cap;
} FILE_LIST;

static inline void fl_free(FILE_LIST *l) {
    size_t i = 0;
    while (i < l->size) {
        free((void *)l->out_paths[i++]);
    }
    free(l->out_paths);
    free(l);
}

static inline FILE_LIST *fl_new() {
    FILE_LIST *new = malloc(sizeof(FILE_LIST));
    new->out_paths = NULL;
    new->size = 0;
    new->cap = 0;
    return new;
}

static inline void fl_incert(FILE_LIST *l, const char *entry) {
    if (l->cap == 0) {
        l->cap = 8;
        l->out_paths = malloc(sizeof(const char *) * l->cap);
    } else if (l->size == l->cap) {
        l->cap *= 2;
        l->out_paths = realloc(l->out_paths, sizeof(const char *) * l->cap);
    }
    l->out_paths[l->size++] = entry;
}

typedef struct {
    BD_FILE_H *fp_in;
    FILE *fp_out;
    int64_t wrote;
    int64_t total;

    BLURAY *bd;
    unsigned num;
    FILE_LIST *l;
    size_t out_len;
} DUMP_PROGRESS;

static inline void progress_free(DUMP_PROGRESS *prog) {
    if (prog->fp_in) file_close(prog->fp_in);
    if (prog->fp_out) fclose(prog->fp_out);
    if (prog->bd) bd_close(prog->bd);
    fl_free(prog->l);
    free(prog);
}

static inline DUMP_PROGRESS *
progress_new(const char *disc_dir, const char *key_file, const char *out_dir) {
    DUMP_PROGRESS *new = malloc(sizeof(DUMP_PROGRESS));
    new->fp_in = NULL;
    new->fp_out = NULL;
    new->wrote = 0;
    new->total = 0;

    new->num = 0;
    new->l = fl_new();
    new->out_len =
        (out_dir ? strlen(out_dir) : strlen(DEFAULT_OUT)) + strlen(DIR_SEP);

    /* open disc */
    new->bd = bd_open(disc_dir, key_file);
    if (!new->bd) {
        fprintf(stderr, "Error opening %s\n", disc_dir);
        progress_free(new);
        new = NULL;
    } else {
        free((void *)disc_dir);
        free((void *)key_file);
    }
    return new;
}

void walk_dir(BLURAY *bd, const char *dir, const char *out_dir,
              DUMP_PROGRESS *prog) {
    BD_DIRENT ep;
    BD_DIR_H *dp = (BD_DIR_H *)bd_open_dir(bd, dir);

    if (dp) {
        if (!dir_read(dp, &ep)) { // non-empty dir
            char *out = malloc(prog->out_len + strlen(DIR_SEP) + strlen(dir) +
                               strlen(DIR_SEP) + NAME_MAX + 1);
            char *tail = stpcpy(
                stpcpy(stpcpy(out, out_dir ? out_dir : DEFAULT_OUT), DIR_SEP),
                dir);
            char *rel = out + prog->out_len;
            mkdir(out, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH); // drwxrwxr-x
            do {
                const char *name = ep.d_name;
                if (!(strcmp(name, ".") && strcmp(name, ".."))) {
                    continue;
                }
                strcpy(*dir ? stpcpy(tail, DIR_SEP) : tail, name);
                if (strstr(name, ".")) {
                    fl_incert(prog->l, strdup(out));
                } else {
                    walk_dir(bd, rel, out_dir, prog);
                }
            } while (!dir_read(dp, &ep));
            free(out);
        }
        dir_close(dp);
    } else {
        fprintf(stderr, "failed to open %s\n", dir);
    }
}

gboolean dump_chunk(gpointer data) {
    DUMP_PROGRESS *prog = data;
    const char *out_path = prog->l->out_paths[prog->num];
    const char *in_path = out_path + prog->out_len;

    while (!prog->fp_out) {
        if (prog->num == prog->l->size) {
            progress_free(prog);
            printf("Done.\n");
            return G_SOURCE_REMOVE;
        }

        /* open input file */
        prog->fp_in = bd_open_file_dec(prog->bd, in_path);
        if (!prog->fp_in) {
            fprintf(stderr, "failed to open %s\n", in_path);
            out_path = prog->l->out_paths[++prog->num];
            in_path = out_path + prog->out_len;
            continue;
        }

        /* open output file */
        prog->fp_out = fopen(out_path, "wb");
        if (!prog->fp_out) {
            fprintf(stderr, "failed to create cache file %s\n", out_path);
            file_close(prog->fp_in);
            prog->fp_in = NULL;
            out_path = prog->l->out_paths[++prog->num];
            in_path = out_path + prog->out_len;
        }
    }

    const size_t buf_size = 6144 * 60;
    uint8_t buf[buf_size];
    int64_t got = file_read(prog->fp_in, buf, buf_size);

    /* we'll call write(fp, buf, 0) after EOF. It is used to check for
     * errors.
     */
    if (got < 0 || fwrite(buf, 1, got, prog->fp_out) != got) {
        fprintf(stderr, "failed to cache file %s\n", out_path);
        remove(out_path);
        prog->wrote = 0;
    } else if (got) {
        prog->wrote += got;
        fprintf(stdout, "%ld MiB.\r",
                (prog->wrote + prog->total) / 1024 / 1024);
        return G_SOURCE_CONTINUE;
    }
    prog->total += prog->wrote;
    prog->wrote = 0;
    fprintf(stdout, "cached %s, total %ld MiB.\n", out_path,
            prog->total / 1024 / 1024);
    file_close(prog->fp_in);
    prog->fp_in = NULL;
    fclose(prog->fp_out);
    prog->fp_out = NULL;
    ++prog->num;
    return G_SOURCE_CONTINUE;
}

typedef struct {
    GtkWidget *keyfile_fileselect;
    GtkWidget *input_folderselect;
    GtkWidget *output_folderselect;
} app_widgets;

void on_execute_clicked(GtkButton *button, app_widgets *widgets) {
    const char *disc_dir = (const char *)gtk_file_chooser_get_filename(
        (GtkFileChooser *)widgets->input_folderselect);
    if (!disc_dir) return;
    const char *key_file = (const char *)gtk_file_chooser_get_filename(
        (GtkFileChooser *)widgets->keyfile_fileselect);
    const char *out_dir = (const char *)gtk_file_chooser_get_filename(
        (GtkFileChooser *)widgets->output_folderselect);

    DUMP_PROGRESS *prog = progress_new(disc_dir, key_file, out_dir);

    /* dump */
    walk_dir(prog->bd, "", out_dir, prog);
    g_idle_add(dump_chunk, prog);
}

// called when window is closed
void on_window_destroy() { gtk_main_quit(); }

int main(int argc, char **argv) {
    GtkBuilder *builder;
    GtkWidget *window;
    app_widgets *widgets = g_slice_new(app_widgets);

    gtk_init(&argc, &argv);
    builder = gtk_builder_new();
    gtk_builder_add_from_file(builder, "blu-save.glade", NULL);

    window = GTK_WIDGET(gtk_builder_get_object(builder, "window"));
    widgets->keyfile_fileselect =
        GTK_WIDGET(gtk_builder_get_object(builder, "keyfile_fileselect"));
    widgets->input_folderselect =
        GTK_WIDGET(gtk_builder_get_object(builder, "input_folderselect"));
    widgets->output_folderselect =
        GTK_WIDGET(gtk_builder_get_object(builder, "output_folderselect"));
    gtk_builder_connect_signals(builder, widgets);
    g_object_unref(builder);

    gtk_widget_show(window);
    gtk_main();
    g_slice_free(app_widgets, widgets);
}
